/*----------------------------------------------------------------------
  File    : xlvq.c
  Contents: learning vector quantization visualization program
  Author  : Christian Borgelt
  History : 2000.05.17 file created from file xbcview.c
            2000.06.09 data display part completed
            2000.06.11 first version completed
            2000.09.01 adapted to module lvq
            2000.09.02 loading reference vectors added
            2001.07.23 adapted to modified module tabread
            2001.11.08 bug in function dcb_data removed
            2003.09.02 drawing of connecting lines added
            2013.08.09 adapted to higher compiler warning level
            2014.10.24 changed from LGPL license to MIT license
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <assert.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Core.h>
#include <X11/Shell.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Viewport.h>
#include <X11/Xaw/List.h>
#include <X11/Xaw/AsciiText.h>
#include "menu.h"
#include "dialog.h"
#include "fselect.h"
#include "xshades.h"
#include "lvqd.h"

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
/* --- extensions --- */
#define MINXEXT      120        /* minimal x-extension of window */
#define MINYEXT       90        /* minimal y-extension of window */
#define INITXEXT     320        /* initial x-extension of window */
#define INITYEXT     320        /* initial y-extension of window */
#define LISTWD       150        /* minimal list width */
#define LISTHT       120        /* minimal list width */

/* --- error codes --- */
#define E_METCNT     (-5)       /* no metric attributes found */
#define E_DATA       (-8)       /* no data loaded */
#define E_INIT       (-9)       /* initialization failed */
#define E_WIDGET    (-10)       /* widget creation failed */
#define E_DIALOG    (-11)       /* cannot create dialog box */
#define E_COLOR     (-12)       /* color allocation failed */
#define E_GRAPH     (-13)       /* cannot create graphics context */
#define E_DLOAD     (-14)       /* loading data failed */

#define FMT_MAX      256        /* max. len. of error message format */

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- coordinate transformation --- */
  double xoff, yoff;            /* offsets */
  double xscl, yscl;            /* scaling factors */
} TRANSFORM;                    /* (coordinate transformation) */

typedef struct {                /* --- error message data --- */
  const char *name, *type;      /* name and type of the error message */
  const char *class;            /* resource class of error message */
  const char *dflt;             /* default message */
  int        argc;              /* number of arguments */
} ERRMSG;                       /* (error message) */

/*----------------------------------------------------------------------
  Action Function Prototypes
----------------------------------------------------------------------*/
static void data   (Widget w, XEvent *e, String *s, Cardinal *c);
static void format (Widget w, XEvent *e, String *s, Cardinal *c);
static void save   (Widget w, XEvent *e, String *s, Cardinal *c);
static void load   (Widget w, XEvent *e, String *s, Cardinal *c);
static void quit   (Widget w, XEvent *e, String *s, Cardinal *c);
static void redraw (Widget w, XEvent *e, String *s, Cardinal *c);
static void init   (Widget w, XEvent *e, String *s, Cardinal *c);
static void exec   (Widget w, XEvent *e, String *s, Cardinal *c);
static void params (Widget w, XEvent *e, String *s, Cardinal *c);
static void colors (Widget w, XEvent *e, String *s, Cardinal *c);
static void atts   (Widget w, XEvent *e, String *s, Cardinal *c);
static void about  (Widget w, XEvent *e, String *s, Cardinal *c);
static void dclick (Widget w, XEvent *e, String *s, Cardinal *c);

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
#include "xlvq.rsc"             /* fallback resources */

static XtActionsRec actions[] = {
  { "data",     data     },     /* File     > Load Data... */
  { "format",   format   },     /* File     > Data Format... */
  { "save",     save     },     /* File     > Save Vectors... */
  { "load",     load     },     /* File     > Load Vectors... */
  { "quit",     quit     },     /* File     > Quit */
  { "redraw",   redraw   },     /* Actions  > Redraw */
  { "init",     init     },     /* Actions  > Initialize */
  { "exec",     exec     },     /* Actions  > Start/Stop Training */
  { "params",   params   },     /* Settings > Parameters... */
  { "colors",   colors   },     /* Settings > Colors... */
  { "atts",     atts     },     /* Settings > Attributes... */
  { "about",    about    },     /* Help     > About XLVQ... */
  { "dclick",   dclick   },     /* double click in attribute list */
  { "db_focus", db_focus },     /* set focus to input gadget */
  { "db_next",  db_next  },     /* switch to next input gadget */
  { "db_close", db_close } };   /* close dialog box */

/* --- error messages --- */
static const ERRMSG errmsgs[] = {   /* error message data */
  /* E_NONE      0 */  { "error",   "none",    "XLVQ.Error",
                         "no error",                               0 },
  /* E_NOMEM    -1 */  { "error",   "nomem",   "XLVQ.Error",
                         "not enough memory",                      0 },
  /* E_FOPEN    -2 */  { "warning", "fopen",   "XLVQ.Warning",
                         "cannot open file:\n%s",                  1 },
  /* E_FREAD    -3 */  { "warning", "fread",   "XLVQ.Warning",
                         "read error on file:\n%s",                1 },
  /* E_FWRITE   -4 */  { "warning", "fwrite",  "XLVQ.Warning",
                         "write error on file:\n%s",               1 },
  /* E_METCNT   -5 */  { "warning", "metcnt",  "XLVQ.Warning",
                         "no metric attribute",                    0 },
  /* E_REFVEC   -6 */  { "warning", "refvec",  "XLVQ.Warning",
                         "no reference vectors",                   0 },
  /* E_RVINC    -7 */  { "warning", "data",    "XLVQ.Warning",
                         "reference vectors are incompatible\n"
                           "with the loaded data",                 0 },
  /* E_DATA     -8 */  { "warning", "data",    "XLVQ.Warning",
                         "no data loaded",                         0 },
  /* E_INIT     -9 */  { "error",   "init",    "XLVQ.Error",
                         "initialization failed",                  0 },
  /* E_WIDGET  -10 */  { "error",   "widget",  "XLVQ.Error",
                         "widget creation failed",                 0 },
  /* E_DIALOG  -11 */  { "warning", "dialog",  "XLVQ.Warning",
                         "cannot create dialog box",               0 },
  /* E_COLOR   -12 */  { "warning", "color",   "XLVQ.Warning",
                         "color allocation failed",                0 },
  /* E_GRAPH   -13 */  { "error",   "gc",      "XLVQ.Error",
                         "cannot create graphics context",         0 },
  /* E_DLOAD   -14 */  { "warning", "data",    "XLVQ.Error",
                         "error loading data:\n%s",                1 },
  /*           -15 */  { "error", "unknown",   "XLVQ.Error",
                         "unknown error",                          0 }};

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static XtAppContext appctx;     /* X11 application context */
static Display *display;        /* display connected to */
static int     screen;          /* screen of the display */
static Window  window;          /* main window */
static Atom    wm_delwin;       /* delete window atom of window mgr. */
static char    fmt[FMT_MAX];    /* format and buffer for messages */
static char    buf[PATH_MAX +FMT_MAX];

/* --- main widgets and menu --- */
static Widget  w_top;           /* top level widget */
static Widget  w_main;          /* main window widget */
static Widget  w_view;          /* viewport widget */
static MENU    *menu;           /* pull down menu */
static GC      gc;              /* graphics context for drawing */

/* --- property sheets --- */
static PSHEET  *psh_fmt    = NULL;  /* data format */
static PSHEET  *psh_params = NULL;  /* learning parameters */
static PSHEET  *psh_color  = NULL;  /* color settings */

/* --- dialog box widgets --- */
static Widget  db_atts = NULL;  /* 'Attributes...' dialog box */
static Widget  di_h_att;        /* attribute list for horizontal */
static Widget  di_h_min;        /* input gadgets for minimal */
static Widget  di_h_max;        /* and maximal horizontal value */
static Widget  di_v_att;        /* attribute list for vertical */
static Widget  di_v_min;        /* input gadgets for minimal */
static Widget  di_v_max;        /* and maximal vertical value */
static Widget  di_c_att;        /* attribute list for class */

/* --- color information --- */
static PIXEL   black;           /* black and white */
static PIXEL   white;           /* for normal drawing */
static SHADES  *shades = NULL;  /* color shades for reference vectors */

static struct {                 /* -- color information -- */
  int    bkgnd;                 /* background color (black or white) */
  double coloff;                /* color offset (in degrees) */
  int    colcnt;                /* number of colors */
  int    shdcnt;                /* number of shades */
  int    size;                  /* size of markers */
} colinfo = { SHD_WHITE, 0.0, 0, 5, 4 };

/* --- training control variables --- */
static int running = 0;         /* whether training is running */
static XtIntervalId iid;        /* id of time interval procedure */

/*----------------------------------------------------------------------
  Error Reporting Functions
----------------------------------------------------------------------*/

static void dcb_nop (Widget widget, XtPointer client, XtPointer call)
{                               /* --- close dialog box (callback) */
  while (!XtIsTransientShell(widget))
    widget = XtParent(widget);  /* find the shell widget */
  XtPopdown(widget);            /* close dialog box */
  mn_enable(menu, MN_MENU, 1);  /* and enable menu bar */
}  /* dcb_nop() */

/*--------------------------------------------------------------------*/

static void error (int code, ...)
{                               /* --- print error message */
  va_list      va;              /* list of variable arguments */
  int          k;               /* minimal error code */
  const ERRMSG *msg;            /* error message */
  String       argv[8];         /* variable argument vector */
  Cardinal     argc;            /* number of arguments */

  k = 1-(int)(sizeof(errmsgs)/sizeof(*errmsgs));
  if (code < k) code = k;       /* check error code against minimum */
  if (code > 0) code = E_DLOAD; /* check for a data load error */
  msg = errmsgs -code;          /* get error message data */
  if (!msg->name) msg = errmsgs -k;
  va_start(va, code);           /* get list of variable arguments */
  for (argc = 0; (int)argc < msg->argc; argc++)
    argv[argc] = va_arg(va, String);
  va_end(va);                   /* get message arguments */
  if (strcmp(msg->name, "error") == 0) /* if name is 'error' */
    XtAppErrorMsg(appctx, msg->name, msg->type, msg->class,
      msg->dflt, argv, &argc);  /* print error message */
  else {                        /* if name is 'warning' */
    XtAppGetErrorDatabaseText(appctx, msg->name, msg->type,
      msg->class, msg->dflt, fmt, FMT_MAX, NULL);
    va_start(va, code);         /* list of variable arguments */
    vsprintf(buf, fmt, va);     /* get and format message */
    va_end(va);                 /* end variable argument evaluation */
    mn_enable(menu, MN_MENU,0); /* disable menu bar and show dialog */
    if (db_alert(w_top, dcb_nop, buf) != 0) {
      mn_enable(menu, MN_MENU, 1);  /* enable menu bar */
      XtAppWarningMsg(appctx, msg->name, msg->type, msg->class,
        msg->dflt, argv, &argc);/* if the alert dialog box */
    }                           /* cannot be shown, */
  }                             /* print a warning message */
}  /* error() */

/*----------------------------------------------------------------------
  Auxiliary Functions
----------------------------------------------------------------------*/

void get_colors (void)
{                               /* --- get color shades */
  if (!attset) return;          /* check for an attribute set */
  if (shades) {                 /* delete existing color shades */
    shd_delete(shades); shades = NULL; }
  colinfo.colcnt = (selatt.c_att < 0) ? 1
                 : att_valcnt(as_att(attset, map_nom[selatt.c_att]));
                                /* get the number of colors */
  colinfo.shdcnt = (colinfo.bkgnd == SHD_BLACK) ? 3 : 5;
  shades = shd_create(display, DefaultColormap(display, screen),
                      colinfo.coloff, colinfo.colcnt,
                      colinfo.bkgnd,  colinfo.shdcnt);
  if (!shades) error(E_COLOR);  /* create color shades */
}  /* get_colors() */

/*--------------------------------------------------------------------*/

static void draw_vecs (TRANSFORM *tf)
{                               /* --- draw reference vectors */
  int       i, k, s;            /* loop variables */
  unsigned  n;                  /* marker size */
  int       x, y, ox, oy;       /* pixel coordinates */
  XGCValues v;                  /* graphic context values */
  Pixel     fill;               /* the fill color */

  assert(tf && shades);         /* check the function arguments */
  if (!rvecs) return;           /* check for existing vectors */
  ox = oy = 0;                  /* clear the coordinates */
  v.line_width = (colinfo.size >> 2) +1;
  XChangeGC(display, gc, GCLineWidth, &v);
  s = (colinfo.size > 1) ? colinfo.size : 2;
  n = (unsigned)((s << 1) +1);  /* get the marker size */
  if (hist) {                   /* if there is a vector history */
    for (i = 0; i < rvcnt; i++){/* traverse the reference vectors */
      XSetForeground(display, gc, shd_pixel(shades, rvcls[i], 1));
      for (k = 0; k < histcnt; k++) {   /* traverse the vectors */
        x = (int)((hist[k][i].x -tf->xoff) *tf->xscl +0.5);
        y = (int)((tf->yoff -hist[k][i].y) *tf->yscl +0.5);
        if (k > 0)              /* draw a connecting line */
          XDrawLine(display, window, gc, ox, oy, x, y);
        XDrawArc(display, window, gc, x-s, y-s, n, n, 0, 360*64);
        ox = x; oy = y;         /* draw a small circle */
      }                         /* for each reference vector */
    }                           /* and note the coordinates */
  }
  fill = (colinfo.bkgnd == SHD_WHITE)
       ? black : white;         /* get the fill color */
  for (i = 0; i < rvcnt; i++) { /* traverse the reference vectors */
    x = (int)((rvecs[i].x -tf->xoff) *tf->xscl +0.5);
    y = (int)((tf->yoff -rvecs[i].y) *tf->yscl +0.5);
    if (hist) {                 /* if there is a vector history */
      XSetForeground(display, gc, shd_pixel(shades, rvcls[i], 1));
      ox = (int)((hist[histcnt-1][i].x -tf->xoff) *tf->xscl +0.5);
      oy = (int)((tf->yoff -hist[histcnt-1][i].y) *tf->yscl +0.5);
      XDrawLine(display, window, gc, ox, oy, x, y);
    }                           /* draw a connecting line */
    x -= s; y -= s;             /* compute coordinates for circle */
    XSetForeground(display, gc, fill);
    XFillArc(display, window, gc, x, y, n, n, 0, 360*64);
    XSetForeground(display, gc, /* set full chroma color */
      shd_pixel(shades, rvcls[i], shd_shdcnt(shades) -1));
    XDrawArc(display, window, gc, x, y, n, n, 0, 360*64);
  }                             /* draw a small circle for each vec. */
  v.line_width = 1;             /* reset the line width and */
  v.cap_style  = CapRound;      /* the cap style (X bug correction) */
  XChangeGC(display, gc, GCLineWidth|GCCapStyle, &v);
}  /* draw_vecs() */

/*----------------------------------------------------------------------
  Neural Network Training Function
----------------------------------------------------------------------*/

static void train (XtPointer data, XtIntervalId *id)
{                               /* --- do one training step */
  Dimension wd, ht;             /* window extensions */
  TRANSFORM tf;                 /* coordinate transformation */

  if (!running) return;         /* check whether training is running */
  lvq_train();                  /* do one training step */
  XtVaGetValues(w_view, XtNwidth, &wd, XtNheight, &ht, NULL);
  tf.xoff = ranges[selatt.h_att].min; /* get coordinate offsets and */
  tf.yoff = ranges[selatt.v_att].max; /* compute the scaling factors */
  tf.xscl = (wd -1) /(ranges[selatt.h_att].max -tf.xoff);
  tf.yscl = (ht -1) /(tf.yoff -ranges[selatt.v_att].min);
  draw_vecs(&tf);               /* draw the reference vectors */
  sprintf(buf, "%d", histcnt);  /* show the current epoch */
  mn_info(menu, buf); mn_resize(menu, wd);
  iid = XtAppAddTimeOut(appctx, (unsigned long)(lvqpar.delay *1000),
          train, NULL);         /* reregister interval procedure */
}  /* train() */

/*----------------------------------------------------------------------
  Dialog Box Callback Functions
----------------------------------------------------------------------*/

static void dcb_data (Widget widget, XtPointer ok, XtPointer fname)
{                               /* --- 'Load Data...' callback */
  int r;                        /* result of load routine */

  mn_enable(menu, MN_MENU, 1);  /* enable menu bar */
  if (!ok) return;              /* if not ok pressed, abort */
  lvq_clean();                  /* clean up reference vectors */
  r = tv_load(fname, 0.1);      /* load a data table for l.v.q. */
  if (r) { error(r, (r < 0) ? fname : tv_msg); return; }
  if (metcnt <= 0) { error(E_METCNT); return; }
  get_colors();                 /* create color shades */
  redraw(NULL,NULL,NULL,NULL);  /* redraw the window contents */
}  /* dcb_data() */

/*--------------------------------------------------------------------*/

static void dcb_format (Widget widget, XtPointer ok, XtPointer call)
{                               /* --- 'Data Format...' callback */
  mn_enable(menu, MN_MENU, 1);  /* enable menu bar */
  if (!ok) return;              /* if not ok pressed, abort */

  datafmt.names = (int)(ptrdiff_t)psh_getval(psh_fmt, "names");
  strncpy(datafmt.blanks,  psh_getval(psh_fmt, "blank"),   FMTCHRLEN);
  datafmt.blanks [FMTCHRLEN] = '\0';  /* get blank characters */
  strncpy(datafmt.fldseps, psh_getval(psh_fmt, "field"),   FMTCHRLEN);
  datafmt.fldseps[FMTCHRLEN] = '\0';  /* get field separators */
  strncpy(datafmt.recseps, psh_getval(psh_fmt, "record"),  FMTCHRLEN);
  datafmt.recseps[FMTCHRLEN] = '\0';  /* get record separators */
  strncpy(datafmt.nullchs, psh_getval(psh_fmt, "null"),    FMTCHRLEN);
  datafmt.nullchs[FMTCHRLEN] = '\0';  /* get null value characters */
  strncpy(datafmt.comment, psh_getval(psh_fmt, "comment"), FMTCHRLEN);
  datafmt.comment[FMTCHRLEN] = '\0';  /* get comment characters */
}  /* dcb_format() */

/*--------------------------------------------------------------------*/

static void dcb_save (Widget widget, XtPointer ok, XtPointer fname)
{                               /* --- 'Save Vectors...' callback */
  mn_enable(menu, MN_MENU, 1);  /* enable menu bar */
  if (!ok) return;              /* if not ok pressed, abort */
  if (lvq_save(fname) != 0)     /* save reference vectors */
    error(E_FWRITE, fname);     /* and check for an error */
}  /* dcb_save() */

/*--------------------------------------------------------------------*/

static void dcb_load (Widget widget, XtPointer ok, XtPointer fname)
{                               /* --- 'Load Vectors...' callback */
  int r;                        /* result of load function */

  mn_enable(menu, MN_MENU, 1);  /* enable menu bar */
  if (!ok) return;              /* if not ok pressed, abort */
  r = lvq_load(fname);          /* load reference vectors */
  if (r < 0) error(r, fname);   /* and check for an error */
  redraw(NULL,NULL,NULL,NULL);  /* redraw the window contents */
}  /* dcb_load() */

/*--------------------------------------------------------------------*/

static void dcb_params (Widget widget, XtPointer ok, XtPointer call)
{                               /* --- 'Parameters...' callback */
  mn_enable(menu, MN_MENU, 1);  /* enable menu bar */
  if (!ok) return;              /* if not ok pressed, abort */
  lvqpar.imode = (int)(ptrdiff_t)psh_getval(psh_params, "imode");
                                /* vector initialization mode */
  lvqpar.vpcls = atoi(psh_getval(psh_params, "vpcls"));
  if      (lvqpar.vpcls < 1)     lvqpar.vpcls = 1;
  else if (lvqpar.vpcls > 100)   lvqpar.vpcls = 100;
                                /* number of vectors per class */
  lvqpar.roff  = atof(psh_getval(psh_params, "roff"));
  if      (lvqpar.roff  < 0)     lvqpar.roff  = 0;
  else if (lvqpar.roff  > 100)   lvqpar.roff  = 10;
                                /* random offset range */
  lvqpar.lrate = atof(psh_getval(psh_params, "lrate"));
  if      (lvqpar.lrate < -100)  lvqpar.lrate = -100;
  else if (lvqpar.lrate > 10)    lvqpar.lrate = 10;
                                /* initial learning rate */
  lvqpar.decay = atof(psh_getval(psh_params, "decay"));
  if      (lvqpar.decay > 1)     lvqpar.decay = 1;
                                /* learning rate decay parameter */
  lvqpar.delay = atof(psh_getval(psh_params, "delay"));
  if      (lvqpar.delay < 0.001) lvqpar.delay = 0.001;
  else if (lvqpar.delay > 10)    lvqpar.delay = 10;
                                /* delay between epochs */
  redraw(NULL,NULL,NULL,NULL);  /* redraw the window contents */
}  /* dcb_params() */

/*--------------------------------------------------------------------*/

static void dcb_colors (Widget widget, XtPointer ok, XtPointer call)
{                               /* --- 'Colors...' callback */
  char buf[32];                 /* temporary buffer */

  mn_enable(menu, MN_MENU, 1);  /* enable menu bar */
  if (!ok) return;              /* if not ok pressed, abort */

  colinfo.bkgnd  = (psh_getval(psh_color, "bkgnd") == (void*)0)
                 ? SHD_WHITE : SHD_BLACK;  /* get background color */
  strncpy(buf, psh_getval(psh_color, "coloff"), 32); buf[31] = '\0';
  colinfo.coloff = atof(buf);   /* get color offset (angle) */
  strncpy(buf, psh_getval(psh_color, "size"),   32); buf[31] = '\0';
  colinfo.size   = atoi(buf);   /* get the marker size */
  if (colinfo.size <= 0) colinfo.size = 1;
  get_colors();                 /* create color shades and */
  redraw(NULL,NULL,NULL,NULL);  /* redraw the window contents */
}  /* dcb_colors() */

/*--------------------------------------------------------------------*/

static void dcb_atts (Widget widget, XtPointer ok, XtPointer call)
{                               /* --- 'Attributes...' callback */
  int    i;                     /* list index */
  String s;                     /* text in input gadgets */

  XtPopdown(db_atts);           /* close dialog box */
  mn_enable(menu, MN_MENU, 1);  /* enable menu bar */
  if (!ok) return;              /* if not ok pressed, abort */

  i = XawListShowCurrent(di_h_att)->list_index;
  if (i != XAW_LIST_NONE) {     /* if an attribute is selected, */
    selatt.h_att = i;           /* note it for the horizontal dir. */
    XtVaGetValues(di_h_min, XtNstring, &s, NULL);
    ranges[selatt.h_att].min = selatt.h_rng.min = atof(s);
    XtVaGetValues(di_h_max, XtNstring, &s, NULL);
    ranges[selatt.h_att].max = selatt.h_rng.max = atof(s);
  }                             /* get the range of values */
  i = XawListShowCurrent(di_v_att)->list_index;
  if (i != XAW_LIST_NONE) {     /* if an attribute is selected, */
    selatt.v_att = i;           /* note it for the vertical dir. */
    XtVaGetValues(di_v_min, XtNstring, &s, NULL);
    ranges[selatt.v_att].min = selatt.v_rng.min = atof(s);
    XtVaGetValues(di_v_max, XtNstring, &s, NULL);
    ranges[selatt.v_att].max = selatt.v_rng.max = atof(s);
  }                             /* get the range of values */
  i = XawListShowCurrent(di_c_att)->list_index;
  selatt.c_att = ((i == XAW_LIST_NONE) || (i >= nomcnt)) ? -1 : i;
  lvq_clean();                  /* delete reference vectors etc. */
  get_colors();                 /* and create new color shades */
  redraw(NULL,NULL,NULL,NULL);  /* redraw the window contents */
}  /* dcb_atts() */

/*----------------------------------------------------------------------
  Menu Callback Functions
----------------------------------------------------------------------*/

static void mcb_data   (Widget widget, XtPointer client, XtPointer call)
{ data(NULL, NULL, NULL, NULL); }

static void mcb_format (Widget widget, XtPointer client, XtPointer call)
{ format(NULL, NULL, NULL, NULL); }

static void mcb_save   (Widget widget, XtPointer client, XtPointer call)
{ save(NULL, NULL, NULL, NULL); }

static void mcb_load   (Widget widget, XtPointer client, XtPointer call)
{ load(NULL, NULL, NULL, NULL); }

static void mcb_quit   (Widget widget, XtPointer client, XtPointer call)
{ quit(NULL, NULL, NULL, NULL); }

static void mcb_redraw (Widget widget, XtPointer client, XtPointer call)
{ redraw(NULL, NULL, NULL, NULL); }

static void mcb_init   (Widget widget, XtPointer client, XtPointer call)
{ init(NULL, NULL, NULL, NULL); }

static void mcb_start  (Widget widget, XtPointer client, XtPointer call)
{ exec(NULL, NULL, NULL, NULL); }

static void mcb_params (Widget widget, XtPointer client, XtPointer call)
{ params(NULL, NULL, NULL, NULL); }

static void mcb_colors (Widget widget, XtPointer client, XtPointer call)
{ colors(NULL, NULL, NULL, NULL); }

static void mcb_atts   (Widget widget, XtPointer client, XtPointer call)
{ atts(NULL, NULL, NULL, NULL); }

static void mcb_about  (Widget widget, XtPointer client, XtPointer call)
{ about(NULL, NULL, NULL, NULL); }

/*----------------------------------------------------------------------
  List Callback Function
----------------------------------------------------------------------*/

static void lcb_att (Widget widget, XtPointer client, XtPointer call)
{                               /* --- attribute list callback */
  int  i;                       /* list index */
  char buf[32];                 /* conversion buffer */

  i = ((XawListReturnStruct*)call)->list_index;
  if ((i == XAW_LIST_NONE) || !ranges)
    return;                     /* check selected list item */
  widget = ((int)(ptrdiff_t)client == 0) ? di_h_min : di_v_min;
  sprintf(buf, "%g", ranges[i].min);
  XtVaSetValues(widget, XtNstring, buf, NULL);
  widget = ((int)(ptrdiff_t)client == 0) ? di_h_max : di_v_max;
  sprintf(buf, "%g", ranges[i].max);
  XtVaSetValues(widget, XtNstring, buf, NULL);
}  /* lcb_att() */              /* set new range of values */

/*----------------------------------------------------------------------
  Action Functions
----------------------------------------------------------------------*/

static void data (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- File > Load Data... */
  const char *t;                /* buffer for dialog box title */

  mn_enable(menu, MN_MENU, 0);  /* disable the menu bar */
  XtVaGetValues(XtNameToWidget(mn_menubar(menu), "*data"),
    XtNlabel, &t, NULL);        /* get the name of the menu entry */
  if (fselect(w_top, t, NULL, NULL, dcb_data) != 0) {
    error(E_DIALOG); mn_enable(menu, MN_MENU, 1); }
}  /* data() */                 /* open file select dialog */

/*--------------------------------------------------------------------*/

static void format (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- File > Data Format... */
  if (!psh_fmt) {               /* if property sheet not created yet */
    psh_fmt = psh_create("fmtDlg", w_top, 1);
    if (!psh_fmt) return;       /* create a property sheet */
    if ((psh_addline(psh_fmt, "lnames")               != 0)
    ||  (psh_additem(psh_fmt, "names",   PSH_MENU, 0) != 0)
    ||  (psh_menuadd(psh_fmt, "names",   "no")        != 0)
    ||  (psh_menuadd(psh_fmt, "names",   "yes")       != 0)
    ||  (psh_addline(psh_fmt, "lrecord")              != 0)
    ||  (psh_additem(psh_fmt, "record",  PSH_EDIT, 0) != 0)
    ||  (psh_addline(psh_fmt, "lfield")               != 0)
    ||  (psh_additem(psh_fmt, "field",   PSH_EDIT, 0) != 0)
    ||  (psh_addline(psh_fmt, "lblank")               != 0)
    ||  (psh_additem(psh_fmt, "blank",   PSH_EDIT, 0) != 0)
    ||  (psh_addline(psh_fmt, "lnull")                != 0)
    ||  (psh_additem(psh_fmt, "null",    PSH_EDIT, 0) != 0)
    ||  (psh_addline(psh_fmt, "lcomment")             != 0)
    ||  (psh_additem(psh_fmt, "comment", PSH_EDIT, 0) != 0)) {
      psh_delete(psh_fmt); error(E_DIALOG); return; }
  }                             /* add dialog items */
  mn_enable(menu, MN_MENU, 0);  /* disable menu bar */
  psh_setval(psh_fmt, "names",   (char*)(ptrdiff_t)datafmt.names);
  psh_setval(psh_fmt, "record",  datafmt.recseps);
  psh_setval(psh_fmt, "field",   datafmt.fldseps);
  psh_setval(psh_fmt, "blank",   datafmt.blanks);
  psh_setval(psh_fmt, "null",    datafmt.nullchs);
  psh_setval(psh_fmt, "comment", datafmt.comment);
  psh_handle(psh_fmt, dcb_format); /* fill input gadgets with current */
}  /* format() */                  /* values and handle dialog */

/*--------------------------------------------------------------------*/

static void save (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- File > Save Vectors... */
  const char *t;                /* buffer for dialog box title */

  if (!rvecs) {                 /* if there are no reference vectors, */
    error(E_REFVEC); return; }  /* abort the function */
  mn_enable(menu, MN_MENU, 0);  /* disable the menu bar */
  XtVaGetValues(XtNameToWidget(mn_menubar(menu), "*save"),
    XtNlabel, &t, NULL);        /* get the name of the menu entry */
  if (fselect(w_top, t, NULL, NULL, dcb_save) != 0) {
    error(E_DIALOG); mn_enable(menu, MN_MENU, 1); }
}  /* save() */                 /* open file select dialog */

/*--------------------------------------------------------------------*/

static void load (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- File > Load Vectors... */
  const char *t;                /* buffer for dialog box title */

  if (!table) {                 /* if there is no data table, */
    error(E_DATA); return; }    /* abort the function */
  mn_enable(menu, MN_MENU, 0);  /* disable the menu bar */
  XtVaGetValues(XtNameToWidget(mn_menubar(menu), "*load"),
    XtNlabel, &t, NULL);        /* get the name of the menu entry */
  if (fselect(w_top, t, NULL, NULL, dcb_load) != 0) {
    error(E_DIALOG); mn_enable(menu, MN_MENU, 1); }
}  /* load() */                 /* open file select dialog */

/*--------------------------------------------------------------------*/

static void quit (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- File > Quit */
  lvq_clean();                  /* clean up reference vectors */
  tv_clean();                   /* and the data table, maps etc. */
  if (shades) shd_delete(shades);
  if (w_top) XtDestroyApplicationContext(appctx);
  exit(0);                      /* terminate the program */
}  /* quit() */

/*--------------------------------------------------------------------*/

static void init (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Actions > Initialize */
  if (!table) { error(E_DATA); return; }
  lvq_init();                   /* initialize the reference vectors */
  redraw(NULL,NULL,NULL,NULL);  /* redraw the window contents */
}  /* init() */

/*--------------------------------------------------------------------*/

static void exec (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Actions > Start/Stop Training */
  if (running) {                /* if running, remove procedure */
    running = 0; XtRemoveTimeOut(iid);
    mn_visible(menu, MN_ITEMS, 3);
    mn_enable (menu, MN_ITEMS, 3); }
  else {                        /* otherwise start training */
    if (!rvecs) init(NULL, NULL, NULL, NULL);
    if (!rvecs) return;
    mn_visible(menu, MN_ITEMS, 5);
    mn_enable (menu, MN_ITEMS, 4);
    running = 1;
    train(NULL, NULL);
  }
}  /* exec() */

/*--------------------------------------------------------------------*/

static void params (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Settings > Parameters... */
  char buf[32];                 /* buffer for numbers */

  if (!psh_params) {            /* if property sheet not created yet */
    psh_params = psh_create("paramDlg", w_top, 1);
    if (!psh_params) return;    /* create a property sheet */
    if ((psh_addline(psh_params, "limode")             != 0)
    ||  (psh_additem(psh_params, "imode", PSH_MENU, 0) != 0)
    ||  (psh_menuadd(psh_params, "imode", "center")    != 0)
    ||  (psh_menuadd(psh_params, "imode", "mean")      != 0)
    ||  (psh_menuadd(psh_params, "imode", "clsmean")   != 0)
    ||  (psh_menuadd(psh_params, "imode", "data")      != 0)
    ||  (psh_addline(psh_params, "lvpcls")             != 0)
    ||  (psh_additem(psh_params, "vpcls", PSH_EDIT, 0) != 0)
    ||  (psh_addline(psh_params, "lroff")              != 0)
    ||  (psh_additem(psh_params, "roff",  PSH_EDIT, 0) != 0)
    ||  (psh_addline(psh_params, "llrate")             != 0)
    ||  (psh_additem(psh_params, "lrate", PSH_EDIT, 0) != 0)
    ||  (psh_addline(psh_params, "ldecay")             != 0)
    ||  (psh_additem(psh_params, "decay", PSH_EDIT, 0) != 0)
    ||  (psh_addline(psh_params, "ldelay")             != 0)
    ||  (psh_additem(psh_params, "delay", PSH_EDIT, 0) != 0)) {
      psh_delete(psh_params); error(E_DIALOG); return; }
  }                             /* set browse button callback */
  mn_enable(menu, MN_MENU, 0);  /* disable menu bar */
  sprintf(buf, "%d", lvqpar.vpcls);      /* vectors per class and */
  psh_setval(psh_params, "vpcls", buf);  /* initialization mode */
  psh_setval(psh_params, "imode", (char*)(ptrdiff_t)lvqpar.imode);
  sprintf(buf, "%g", lvqpar.roff);       /* random offset range */
  psh_setval(psh_params, "roff",  buf);
  sprintf(buf, "%g", lvqpar.lrate);      /* learning rate */
  psh_setval(psh_params, "lrate", buf);
  sprintf(buf, "%g", lvqpar.decay);      /* learning rate */
  psh_setval(psh_params, "decay", buf);  /* decay parameter */
  sprintf(buf, "%g", lvqpar.delay);      /* delay between */
  psh_setval(psh_params, "delay", buf);  /* training steps */
  psh_handle(psh_params, dcb_params);    /* handle the dialog */
}  /* params() */

/*--------------------------------------------------------------------*/

static void dclick (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- double click in attribute list */
  dcb_atts(db_atts, (void*)1, NULL);
}  /* dclick() */               /* call dialog box callback for 'ok' */

/*--------------------------------------------------------------------*/

static int listbox (Widget form, Widget left, const char *name)
{                               /* --- create a list box */
  Cardinal n;                   /* widget argument counter */
  Arg      args[16];            /* widget argument list */
  Widget   view, list;          /* viewport and list widget */
  Widget   min, max;            /* minimum and maximum widget */
  Widget   label;               /* buffer for label widgets */
  char     buf[16];             /* buffer for widget name */

  /* --- create list box --- */
  n = 0;                      /* build argument list */
  XtSetArg(args[n], XtNborderWidth, 0);                       n++;
  if (left) { XtSetArg(args[n], XtNfromHoriz,   left);        n++; }
  else      { XtSetArg(args[n], XtNleft,        XtChainLeft); n++; }
  XtSetArg(args[n], XtNtop,         XtChainTop);              n++;
  XtSetArg(args[n], XtNbottom,      XtChainTop);              n++;
  XtSetArg(args[n], XtNjustify,     XtJustifyLeft);           n++;
  sprintf(buf, "l%s", name);  /* create the widget name */
  label = XtCreateManagedWidget(buf, labelWidgetClass, form, args, n);
  if (!label) return -1;      /* create title widget */
  n = 0;                      /* build argument list */
  XtSetArg(args[n], XtNborderWidth, 1);                       n++;
  if (left) { XtSetArg(args[n], XtNfromHoriz,   left);        n++; }
  else      { XtSetArg(args[n], XtNleft,        XtChainLeft); n++; }
  XtSetArg(args[n], XtNfromVert,    label);                   n++;
  XtSetArg(args[n], XtNallowHoriz,  False);                   n++;
  XtSetArg(args[n], XtNallowVert,   True);                    n++;
  XtSetArg(args[n], XtNuseRight,    True);                    n++;
  XtSetArg(args[n], XtNtop,         XtChainTop);              n++;
  XtSetArg(args[n], XtNbottom,      XtChainBottom);           n++;
  sprintf(buf, "v%s", name);  /* create the widget name */
  view = XtCreateManagedWidget(buf, viewportWidgetClass,
                               form, args, n);
  if (!view) return -1;       /* create list viewport */
  n = 0;                      /* build argument list */
  XtSetArg(args[n], XtNstring,         "");                   n++;
  XtSetArg(args[n], XtNborderWidth,    0);                    n++;
  XtSetArg(args[n], XtNverticalList,   True);                 n++;
  XtSetArg(args[n], XtNdefaultColumns, 1);                    n++;
  XtSetArg(args[n], XtNforceColumns,   True);                 n++;
  sprintf(buf, "%s_att", name);  /* create the widget name */
  list = XtCreateManagedWidget(buf, listWidgetClass, view, args, n);
  if (!list) return -1;       /* create list widget */

  /* --- create input field for minimum --- */
  n = 0;                      /* build argument list */
  XtSetArg(args[n], XtNborderWidth, 0);                       n++;
  XtSetArg(args[n], XtNfromVert,    view);                    n++;
  if (left) { XtSetArg(args[n], XtNfromHoriz,   left);        n++; }
  else      { XtSetArg(args[n], XtNleft,        XtChainLeft); n++; }
  XtSetArg(args[n], XtNtop,         XtChainBottom);           n++;
  XtSetArg(args[n], XtNbottom,      XtChainBottom);           n++;
  XtSetArg(args[n], XtNjustify,     XtJustifyLeft);           n++;
  sprintf(buf, "l%s_min", name); /* create the widget name */
  label = XtCreateManagedWidget(buf, labelWidgetClass, form, args, n);
  if (!label) return -1;      /* create min label widget */
  n = 0;                      /* build argument list */
  XtSetArg(args[n], XtNborderWidth, 1);                       n++;
  XtSetArg(args[n], XtNeditType,    XawtextEdit);             n++;
  XtSetArg(args[n], XtNfromHoriz,   label);                   n++;
  XtSetArg(args[n], XtNfromVert,    view);                    n++;
  XtSetArg(args[n], XtNtop,         XtChainBottom);           n++;
  XtSetArg(args[n], XtNbottom,      XtChainBottom);           n++;
  min = XtCreateManagedWidget(buf+1, asciiTextWidgetClass,
                                 form, args, n);
  if (!min) return -1;        /* create minimum widget */

  /* --- create input field for maximum --- */
  n = 0;                      /* build argument list */
  XtSetArg(args[n], XtNborderWidth, 0);                       n++;
  XtSetArg(args[n], XtNfromVert,    min);                     n++;
  if (left) { XtSetArg(args[n], XtNfromHoriz,   left);        n++; }
  else      { XtSetArg(args[n], XtNleft,        XtChainLeft); n++; }
  XtSetArg(args[n], XtNtop,         XtChainBottom);           n++;
  XtSetArg(args[n], XtNbottom,      XtChainBottom);           n++;
  XtSetArg(args[n], XtNjustify,     XtJustifyLeft);           n++;
  sprintf(buf, "l%s_max", name); /* create the widget name */
  label = XtCreateManagedWidget(buf, labelWidgetClass, form, args, n);
  if (!label) return -1;      /* create h_max label widget */
  n = 0;                      /* build argument list */
  XtSetArg(args[n], XtNborderWidth, 1);                       n++;
  XtSetArg(args[n], XtNeditType,    XawtextEdit);             n++;
  XtSetArg(args[n], XtNfromHoriz,   label);                   n++;
  XtSetArg(args[n], XtNfromVert,    min);                     n++;
  XtSetArg(args[n], XtNtop,         XtChainBottom);           n++;
  XtSetArg(args[n], XtNbottom,      XtChainBottom);           n++;
  max = XtCreateManagedWidget(buf+1, asciiTextWidgetClass,
                              form, args, n);
  if (!max) return -1;          /* create maximum widget */
  return 0;                     /* return 'ok' */
}  /* listbox() */

/*--------------------------------------------------------------------*/

static void atts (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Settings > Attributes... */
  static String empty = "";     /* an empty string */
  XtTranslations tla, tlo;      /* translations for text widgets */
  Widget    shell = NULL, form; /* widgets for dialog shell and form */
  Widget    lview, rview;       /* left and right viewport widget */
  Widget    cview;              /* class viewport widget */
  Widget    label;              /* buffer for label widgets */
  Widget    lh_min, lh_max;     /* label widgets for horizontal */
  Widget    lv_min, lv_max;     /* and vertical minimum/maximum */
  Widget    ok, cancel;         /* widgets for ok and cancel buttons */
  Arg       args[16];           /* widget argument list */
  Cardinal  n;                  /* widget argument counter */
  Position  x,  y;              /* position of created dialog */
  Dimension wd, ht;             /* width and height of created dialog */
  Dimension lblwd, t;           /* label width and temp. buffer */
  int       dd;                 /* default distance of dialog items */

  if (!db_atts) {               /* if dialog box is not created */
    shell = XtCreatePopupShell("attDlg", transientShellWidgetClass,
              w_top, NULL, 0);  /* create a dialog box shell */
    if (!shell) { error(E_DIALOG); return; }
    XtSetArg(args[0], XtNborderWidth, 0);
    form = XtCreateManagedWidget("form", formWidgetClass,
             shell, args, 1);   /* create a form for the box layout */
    if (!form) { XtDestroyWidget(shell); error(E_DIALOG); return; }

     /* --- create listbox for horizontal axis --- */
    if (listbox(form, NULL,  "h") != 0) {
      XtDestroyWidget(shell); error(E_DIALOG); return; }
    lview    = XtNameToWidget(form, "*vh");
    di_h_att = XtNameToWidget(form, "*h_att");
    di_h_min = XtNameToWidget(form, "*h_min");
    di_h_max = XtNameToWidget(form, "*h_max");
    XtAddCallback(di_h_att, XtNcallback, lcb_att, (void*)0);

    /* --- create listbox for vertical axis --- */
    if (listbox(form, lview, "v") != 0) {
      XtDestroyWidget(shell); error(E_DIALOG); return; }
    rview    = XtNameToWidget(form, "*vv");
    di_v_att = XtNameToWidget(form, "*v_att");
    di_v_min = XtNameToWidget(form, "*v_min");
    di_v_max = XtNameToWidget(form, "*v_max");
    XtAddCallback(di_v_att, XtNcallback, lcb_att, (void*)1);

    /* --- create list box for class attribute --- */
    n = 0;                      /* build argument list */
    XtSetArg(args[n], XtNborderWidth, 0);             n++;
    XtSetArg(args[n], XtNfromHoriz,   rview);         n++;
    XtSetArg(args[n], XtNtop,         XtChainTop);    n++;
    XtSetArg(args[n], XtNbottom,      XtChainTop);    n++;
    XtSetArg(args[n], XtNjustify,     XtJustifyLeft); n++;
    label = XtCreateManagedWidget("lclass", labelWidgetClass,
              form, args, n);   /* create title widget */
    if (!label) { XtDestroyWidget(shell); error(E_DIALOG); return; }
    n = 0;                      /* build argument list */
    XtSetArg(args[n], XtNborderWidth, 1);             n++;
    XtSetArg(args[n], XtNfromHoriz,   rview);         n++;
    XtSetArg(args[n], XtNfromVert,    label);         n++;
    XtSetArg(args[n], XtNallowHoriz,  False);         n++;
    XtSetArg(args[n], XtNallowVert,   True);          n++;
    XtSetArg(args[n], XtNuseRight,    True);          n++;
    XtSetArg(args[n], XtNtop,         XtChainTop);    n++;
    XtSetArg(args[n], XtNright,       XtChainRight);  n++;
    XtSetArg(args[n], XtNbottom,      XtChainBottom); n++;
    cview = XtCreateManagedWidget("listView", viewportWidgetClass,
              form, args, n);   /* create list viewport */
    if (!cview) { XtDestroyWidget(shell); error(E_DIALOG); return; }
    n = 0;                      /* build argument list */
    XtSetArg(args[n], XtNstring,         "");         n++;
    XtSetArg(args[n], XtNborderWidth,    0);          n++;
    XtSetArg(args[n], XtNverticalList,   True);       n++;
    XtSetArg(args[n], XtNdefaultColumns, 1);          n++;
    XtSetArg(args[n], XtNforceColumns,   True);       n++;
    di_c_att = XtCreateManagedWidget("class", listWidgetClass,
                 cview, args, n);  /* create list widget */
    if (!di_c_att) { XtDestroyWidget(shell); error(E_DIALOG); return; }

    /* --- create ok button --- */
    n = 0;                      /* build argument list */
    XtSetArg(args[n], XtNfromHoriz,   rview);         n++;
    XtSetArg(args[n], XtNfromVert,    cview);         n++;
    XtSetArg(args[n], XtNright,       XtChainRight);  n++;
    XtSetArg(args[n], XtNtop,         XtChainBottom); n++;
    XtSetArg(args[n], XtNbottom,      XtChainBottom); n++;
    ok = XtCreateManagedWidget("ok", commandWidgetClass,
           form, args, n);      /* create 'ok' button */
    if (!ok) { XtDestroyWidget(shell); error(E_DIALOG); return; }
    XtAddCallback(ok, XtNcallback, dcb_atts, (void*)1);

    /* --- create cancel button --- */
    n = 0;                      /* build argument list */
    XtSetArg(args[n], XtNfromHoriz,   rview);         n++;
    XtSetArg(args[n], XtNfromVert,    ok);            n++;
    XtSetArg(args[n], XtNright,       XtChainRight);  n++;
    XtSetArg(args[n], XtNtop,         XtChainBottom); n++;
    XtSetArg(args[n], XtNbottom,      XtChainBottom); n++;
    cancel = XtCreateManagedWidget("cancel", commandWidgetClass,
               form, args, n);  /* create 'cancel' button */
    if (!cancel) { XtDestroyWidget(shell); error(E_DIALOG); return; }
    XtAddCallback(cancel, XtNcallback, dcb_atts, (void*)0);

    /* --- adapt widths of labels and gadgets --- */
    XtSetArg(args[0], XtNwidth, &t); lblwd = 1;
    lh_min = XtNameToWidget(form, "*lh_min");
    XtGetValues(lh_min, args, 1); if (t > lblwd) lblwd = t;
    lh_max = XtNameToWidget(form, "*lh_max");
    XtGetValues(lh_max, args, 1); if (t > lblwd) lblwd = t;
    lv_min = XtNameToWidget(form, "*lv_min");
    XtGetValues(lv_min, args, 1); if (t > lblwd) lblwd = t;
    lv_max = XtNameToWidget(form, "*lv_max");
    XtGetValues(lv_max, args, 1); if (t > lblwd) lblwd = t;
    XtVaSetValues(lh_min,   XtNwidth, lblwd, NULL);
    XtVaSetValues(lh_max,   XtNwidth, lblwd, NULL);
    XtVaSetValues(lv_min,   XtNwidth, lblwd, NULL);
    XtVaSetValues(lv_max,   XtNwidth, lblwd, NULL);
    XtVaGetValues(di_h_min, XtNwidth, &wd, NULL);
    #ifdef LINUX                /* if Linux Athena Widgets */
    XtVaGetValues(form,     XtNdefaultDistance, &dd, NULL);
    #else                       /* get the default distance, */
    dd = 4;                     /* if Sun or HP Athena Widgets, */
    #endif                      /* set the default distance */
    wd = (Dimension)(wd +lblwd +dd); /* compute the listbox width */
    if (wd < LISTWD) wd = LISTWD;    /* and set it */
    XtVaSetValues(rview,    XtNwidth, wd, NULL);
    XtVaSetValues(lview,    XtNwidth, wd, NULL);
    XtVaSetValues(cview,    XtNwidth, wd, NULL);
    XtVaSetValues(ok,       XtNwidth, wd, NULL);
    XtVaSetValues(cancel,   XtNwidth, wd, NULL);
    wd = (Dimension)(wd -lblwd -dd); /* compute the input field width */
    XtVaSetValues(di_h_min, XtNwidth, wd, NULL);
    XtVaSetValues(di_h_max, XtNwidth, wd, NULL);
    XtVaSetValues(di_v_min, XtNwidth, wd, NULL);
    XtVaSetValues(di_v_max, XtNwidth, wd, NULL);

    /* --- install translations --- */
    tla = XtParseTranslationTable(
            "<Btn1Down>,<Btn1Up>: Set()Notify()\n"
            "<Btn1Down>(2+):      dclick()");
    XtAugmentTranslations(di_h_att, tla);
    XtAugmentTranslations(di_v_att, tla);
    XtAugmentTranslations(di_c_att, tla);
    tla = XtParseTranslationTable(
            "<Key>Home  : beginning-of-line()\n"
            "<Key>End   : end-of-line()\n");
    tlo = XtParseTranslationTable(
            "<Key>Tab   : db_next()\n"
            "<Key>Return: db_close(ok)\n"
            "!Alt<Key>o : db_close(ok)\n"
            "!Alt<Key>c : db_close(cancel)\n"
            "<Btn1Down> : db_focus()\n");
    XtAugmentTranslations (di_h_min, tla);
    XtOverrideTranslations(di_h_min, tlo);
    XtAugmentTranslations (di_h_max, tla);
    XtOverrideTranslations(di_h_max, tlo);
    XtAugmentTranslations (di_v_min, tla);
    XtOverrideTranslations(di_v_min, tlo);
    XtAugmentTranslations (di_v_max, tla);
    XtOverrideTranslations(di_v_max, tlo);

    XtInstallAllAccelerators(form, form);
    db_atts = shell;            /* install actions and accelerators */
  }                             /* and note the created dialog box */
  mn_enable(menu, MN_MENU, 0);  /* disable menu bar */

  /* --- fill dialog box fields --- */
  if (!nms_met || (metcnt <= 0)) { /* if the name list is not valid */
    XawListChange(di_h_att, (const char**)&empty, 1, 0, 1);
    XawListChange(di_v_att, (const char**)&empty, 1, 0, 1);
    XtSetArg(args[0], XtNstring, "");
    XtSetValues(di_h_min, args, 1);  /* clear the list boxes */
    XtSetValues(di_h_max, args, 1);  /* and the input gadgets */
    XtSetValues(di_v_min, args, 1);
    XtSetValues(di_v_max, args, 1); }
  else {                        /* if the name list is valid */
    XawListChange(di_h_att, (const char**)nms_met, metcnt, 0, 1);
    XawListHighlight(di_h_att, selatt.h_att);
    XawListChange(di_v_att, (const char**)nms_met, metcnt, 0, 1);
    XawListHighlight(di_v_att, selatt.v_att);
    sprintf(buf, "%g", selatt.h_rng.min);
    XtVaSetValues(di_h_min, XtNstring, buf, NULL);
    sprintf(buf, "%g", selatt.h_rng.max);
    XtVaSetValues(di_h_max, XtNstring, buf, NULL);
    sprintf(buf, "%g", selatt.v_rng.min);
    XtVaSetValues(di_v_min, XtNstring, buf, NULL);
    sprintf(buf, "%g", selatt.v_rng.max);
    XtVaSetValues(di_v_max, XtNstring, buf, NULL);
  }                             /* set attribute names and ranges */
  if (!nms_nom || (nomcnt <= 0))/* if the name list is not valid */
    XawListChange(di_c_att, (const char**)&empty, 1, 0, 1);
  else {                        /* if the name list is valid */
    XawListChange(di_c_att, (const char**)nms_nom, nomcnt+1, 0, 1);
    if (selatt.c_att >= 0) XawListHighlight(di_c_att, selatt.c_att);
  }                             /* set class attribute names */
  XtVaSetValues(XtParent(di_h_att), XtNheight, LISTHT, NULL);
  XtVaSetValues(XtParent(di_v_att), XtNheight, LISTHT, NULL);
  XtVaSetValues(XtParent(di_c_att), XtNheight, LISTHT, NULL);

  /* --- show dialog box --- */
  XtVaGetValues(w_top,   XtNx, &x, XtNy, &y, NULL);
  XtVaSetValues(db_atts, XtNx,  x, XtNy,  y, NULL);
  XtPopup(db_atts, XtGrabExclusive);   /* place and show dialog box */
  XtSetKeyboardFocus(db_atts, di_h_min);  /* and set keyboard focus */

  if (shell) {                  /* if dialog box has been created */
    XtVaGetValues(shell, XtNwidth,   &wd, XtNheight,   &ht, NULL);
    XtVaSetValues(shell, XtNminWidth, wd, XtNminHeight, ht, NULL);
    XtOverrideTranslations(shell,
      XtParseTranslationTable("<Message>WM_PROTOCOLS: db_close()"));
    XSetWMProtocols(XtDisplay(shell), XtWindow(shell),
      &wm_delwin, 1);           /* override window manager delete */
  }
}  /* atts() */

/*--------------------------------------------------------------------*/

static void colors (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Settings > Colors... */
  char buf[32];                 /* temporary buffer */

  if (!psh_color) {             /* if property sheet not created yet */
    psh_color = psh_create("colorDlg", w_top, 1);
    if (!psh_color) return;     /* create a property sheet */
    if ((psh_addline(psh_color, "lbkgnd")              != 0)
    ||  (psh_additem(psh_color, "bkgnd",  PSH_MENU, 0) != 0)
    ||  (psh_menuadd(psh_color, "bkgnd",  "white")     != 0)
    ||  (psh_menuadd(psh_color, "bkgnd",  "black")     != 0)
    ||  (psh_addline(psh_color, "lcoloff")             != 0)
    ||  (psh_additem(psh_color, "coloff", PSH_EDIT, 0) != 0)
    ||  (psh_addline(psh_color, "lsize")               != 0)
    ||  (psh_additem(psh_color, "size",   PSH_EDIT, 0) != 0)) {
      psh_delete(psh_color); error(E_DIALOG); return; }
  }                             /* add dialog items */
  mn_enable(menu, MN_MENU, 0);  /* disable menu bar */
  psh_setval(psh_color, "bkgnd",         /* background color */
             (char*)(ptrdiff_t)((colinfo.bkgnd == SHD_BLACK) ? 1 : 0));
  sprintf(buf, "%g", colinfo.coloff);
  psh_setval(psh_color, "coloff", buf);  /* color offset */
  sprintf(buf, "%d", colinfo.size);
  psh_setval(psh_color, "size", buf);    /* marker size */
  psh_handle(psh_color, dcb_colors);     /* handle the dialog */
}  /* colors() */

/*--------------------------------------------------------------------*/

static void redraw (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Actions > Redraw */
  static Dimension owd = 0;     /* old window width */
  static Dimension oht = 0;     /* old window height */
  int        i, k;              /* loop variables, buffers */
  unsigned   n;                 /* size/width/height */
  float      f;                 /* temporary buffer */
  TUPLE      *tpl;              /* to traverse the data tuples */
  int        xid, yid, clsid;   /* attribute identifiers */
  int        xtype, ytype;      /* metric attribute types */
  int        cls;               /* current class value */
  double     rx, ry;            /* real  coordinates */
  int        x,  y;             /* pixel coordinates */
  XEvent     de;                /* dummy event */
  XRectangle rc;                /* clipping rectangle */
  Dimension  wd, ht;            /* window extensions */
  TRANSFORM  tf;                /* coordinate transformation */
  PIXEL      pxl, dflt;         /* current and default pixel value */
  int        shdcnt;            /* number of available shades */

  /* --- initialize graphics --- */
  if (!gc) {                    /* if there is no graphics context */
    gc = XCreateGC(display, window, 0, NULL);
    if (!gc) { error(E_GRAPH); return; }
  }                             /* create a graphics context */
  if (!attset || !table || !shades) {
    XClearWindow(display, window);
    return;                     /* if there is no data to draw, */
  }                             /* clear the window and abort */
  XtVaGetValues(w_view, XtNwidth, &wd, XtNheight, &ht, NULL);
  if (e && (e->type == Expose)  /* if the window was exposed, */
  &&  (wd == owd) && (ht == oht)) {       /* but not resized, */
    rc.x      = (short)e->xexpose.x;      /* build and set */
    rc.y      = (short)e->xexpose.y;      /* clipping rectangle */
    rc.width  = (unsigned short)e->xexpose.width;
    rc.height = (unsigned short)e->xexpose.height;
    XSetClipRectangles(display, gc, 0, 0, &rc, 1, Unsorted); }
  else {                        /* if the window was resized */
    rc.x = rc.y = 0;            /* or this is a forced redraw, */
    rc.width  = wd;             /* redraw all of the window */
    rc.height = ht;             /* (set full window rectangle) */
    while (XCheckTypedWindowEvent(display, window, Expose, &de));
  }                             /* remove all exposure events */
  owd = wd; oht = ht;           /* note (new) window dimensions */

  tf.xoff = ranges[selatt.h_att].min; /* get coordinate offsets and */
  tf.yoff = ranges[selatt.v_att].max; /* compute the scaling factors */
  tf.xscl = (wd -1) /(ranges[selatt.h_att].max -tf.xoff);
  tf.yscl = (ht -1) /(tf.yoff -ranges[selatt.v_att].min);

  /* --- draw background --- */
  pxl = (colinfo.bkgnd == SHD_WHITE) ? white : black;
  XSetForeground(display, gc, pxl);
  XFillRectangle(display, window, gc, 0, 0, wd, ht);
  shdcnt = shd_shdcnt(shades);

  /* --- draw data --- */
  clsid = (selatt.c_att < 0)    /* get class attribute identifier */
        ? -1 : map_nom[selatt.c_att];
  xid   = map_met[selatt.h_att];/* get metric attribute ids. */
  yid   = map_met[selatt.v_att];/* and attribute types */
  xtype = att_type(as_att(attset, xid));
  ytype = att_type(as_att(attset, yid));
  dflt  = (colinfo.bkgnd == SHD_WHITE)
        ? black : white;        /* get the default color */
  for (k = tab_tplcnt(table); --k >= 0; ) {
    tpl = tab_tpl(table, k);    /* traverse the data tuples */
    if (xtype == AT_INT) {      /* if x-coordinate is integer valued */
      i = tpl_colval(tpl, xid)->i;
      if (isnull(i)) continue;
      rx = (double)i; }         /* get and check the attribute value */
    else {                      /* if x-coordinate is real valued */
      f = tpl_colval(tpl, xid)->f;
      if (isnan(f))  continue;
      rx = (double)f;           /* get and check the attribute value */
    }
    if (ytype == AT_INT) {      /* if y-coordinate is integer valued */
      i = tpl_colval(tpl, yid)->i;
      if (isnull(i)) continue;
      ry = (double)i; }         /* get and check the attribute value */
    else {                      /* if y-coordinate is real valued */
      f = tpl_colval(tpl, yid)->f;
      if (isnan(f))  continue;
      ry = (double)f;           /* get and check the attribute value */
    }                           /* then compute the screen coords. */
    x   = (int)((rx -tf.xoff) *tf.xscl +0.5);
    y   = (int)((tf.yoff -ry) *tf.yscl +0.5);
    cls = (clsid < 0) ?  -1  : tpl_colval(tpl, clsid)->i;
    pxl = (cls   < 0) ? dflt : shd_pixel(shades, cls, shdcnt-1);
    XSetForeground(display, gc, pxl);
    if (colinfo.size <= 1) {    /* minimal marker size: cross */
      XDrawLine(display, window, gc, x-1, y,   x+1, y);
      XDrawLine(display, window, gc, x,   y-1, x,   y+1); }
    else {                      /* user defined marker size */
      i = colinfo.size -1; x -= i; y -= i; n = (Cardinal)((i << 1) +1);
      XFillArc(display, window, gc, x, y, n, n, 0, 360*64);
      XSetForeground(display, gc, dflt);
      XDrawArc(display, window, gc, x, y, n, n, 0, 360*64);
    }                           /* draw a circle */
  }                             /* at the data point */

  /* --- draw reference vectors --- */
  draw_vecs(&tf);               /* draw vectors and history */
  sprintf(buf, "%d", histcnt);  /* show the current epoch */
  mn_info(menu, buf); mn_resize(menu, wd);

  XFlush(display);              /* flush buffered drawing commands */
  XSetClipMask(display, gc, None);      /* and clear clipping mask */
}  /* redraw() */

/*--------------------------------------------------------------------*/

static void about (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- 'About' dialog box */
  mn_enable(menu, MN_MENU, 0);  /* disable menu bar */
  if (db_about(w_top, 0, dcb_nop) != 0)
    error(E_DIALOG);            /* show 'about' dialog box */
}  /* about() */

/*----------------------------------------------------------------------
  Main Function
----------------------------------------------------------------------*/

int main (int argc, char *argv[])
{                               /* --- main program */
  Arg      args[8];             /* widget arguments */
  Cardinal n;                   /* argument counter */

  XtSetLanguageProc(NULL, (XtLanguageProc)NULL, NULL);

  /* --- create main window --- */
  w_top = XtAppInitialize(&appctx, "XLVQ",
            NULL, 0, &argc, argv, fallback, NULL, 0);
  if (!w_top) error(E_INIT);    /* create toplevel widget */
  n = 0;                        /* build argument list */
  XtSetArg(args[n], XtNdefaultDistance, 0);    n++;
  XtSetArg(args[n], XtNresizable,       True); n++;
  w_main = XtCreateManagedWidget("main", formWidgetClass,
             w_top, args, n);   /* create main widget */
  if (!w_main) error(E_WIDGET); /* (form for window layout) */
  menu = mn_create(w_main);     /* create a menu bar and */
  if (!menu) error(E_NOMEM);    /* add titles and items */
  if ((mn_addtitle(menu, "file",     0, 0,           0) != 0)
  ||  (mn_additem (menu, "data",     1, mcb_data,    0) != 0)
  ||  (mn_additem (menu, "format",   0, mcb_format,  0) != 0)
  ||  (mn_additem (menu, NULL,       0, 0,           0) != 0)
  ||  (mn_additem (menu, "load",     1, mcb_load,    0) != 0)
  ||  (mn_additem (menu, "save",     0, mcb_save,    0) != 0)
  ||  (mn_additem (menu, NULL,       0, 0,           0) != 0)
  ||  (mn_additem (menu, "quit",     0, mcb_quit,    0) != 0)
  ||  (mn_addtitle(menu, "actions",  0, 0,           0) != 0)
  ||  (mn_additem (menu, "init",     0, mcb_init,    0) != 0)
  ||  (mn_additem (menu, "start",    2, mcb_start,   0) != 0)
  ||  (mn_additem (menu, "stop",     4, mcb_start,   0) != 0)
  ||  (mn_additem (menu, NULL,       0, 0,           0) != 0)
  ||  (mn_additem (menu, "redraw",   0, mcb_redraw,  0) != 0)
  ||  (mn_addtitle(menu, "settings", 0, 0,           0) != 0)
  ||  (mn_additem (menu, "params",   0, mcb_params,  0) != 0)
  ||  (mn_additem (menu, "atts",     1, mcb_atts,    0) != 0)
  ||  (mn_additem (menu, "colors",   0, mcb_colors,  0) != 0)
  ||  (mn_addtitle(menu, "help",     0, 0,           0) != 0)
  ||  (mn_additem (menu, "about",    0, mcb_about,   0) != 0)
  ||  (mn_addtitle(menu, NULL,       0, 0,           0) != 0))
    error(E_WIDGET);
  mn_resize (menu, INITXEXT);   /* set width of menu bar */
  mn_visible(menu, MN_ITEMS, 3);/* show/hide      menu items */
  mn_enable (menu, MN_ITEMS, 3);/* enable/disable menu items */

  /* --- create viewport --- */
  n = 0;                        /* build argument list */
  XtSetArg(args[n], XtNfromVert,  mn_menubar(menu)); n++;
  XtSetArg(args[n], XtNleft,      XtChainLeft);      n++;
  XtSetArg(args[n], XtNtop,       XtChainTop);       n++;
  XtSetArg(args[n], XtNright,     XtChainRight);     n++;
  XtSetArg(args[n], XtNbottom,    XtChainBottom);    n++;
  XtSetArg(args[n], XtNresizable, True);             n++;
  w_view = XtCreateManagedWidget("view",
             coreWidgetClass, w_main, args, n);
  if (!w_view) error(E_WIDGET); /* create viewport */
  XtVaSetValues(w_view, XtNwidth,     INITXEXT,
                        XtNheight,    INITYEXT, NULL);
  XtVaSetValues(w_top,  XtNminWidth,  MINXEXT,
                        XtNminHeight, MINYEXT,  NULL);
                                /* set window size */
  XtOverrideTranslations(w_view,/* install redraw function */
    XtParseTranslationTable("<Expose>: redraw()"));

  /* --- install actions, accelerators and translations --- */
  XtAppAddActions(appctx, actions, XtNumber(actions));
  XtInstallAllAccelerators(w_main, w_main);

  /* --- show window --- */
  XtRealizeWidget(w_top);       /* show window on display */
  display = XtDisplay(w_top);   /* get display of top window */
  window  = XtWindow(w_view);   /* and drawable for window */

  if (db_alert(w_top, dcb_nop, NULL) != 0)
    error(E_WIDGET);            /* create alert dialog box */

  /* --- override window manager delete --- */
  XtOverrideTranslations(w_top, /* override window manager delete */
    XtParseTranslationTable("<Message>WM_PROTOCOLS: quit()"));
  wm_delwin = XInternAtom(XtDisplay(w_top), "WM_DELETE_WINDOW", False);
  XSetWMProtocols(XtDisplay(w_top), XtWindow(w_top), &wm_delwin, 1);

  /* --- allocate colors --- */
  screen = DefaultScreen(display);
  black  = BlackPixel(display, screen);
  white  = WhitePixel(display, screen);

  /* --- evaluate command line arguments --- */
  if (argc > 1) dcb_data(NULL, (XtPointer)1, (XtPointer)argv[1]);
  if (argc > 2) dcb_load(NULL, (XtPointer)1, (XtPointer)argv[2]);

  /* --- start event loop --- */
  XtAppMainLoop(appctx);        /* enter main event loop */
  return 0;                     /* return 'ok' */
}  /* main() */
